import torch
import torch.nn as nn
import numpy as np


def load_day_length(path):
    day_length = np.load(path)

    day_id = []
    for i in range(day_length.shape[0]):
        if i == 0:
            day_id.append(day_length[i])
        else:
            append_i = day_id[i - 1] + day_length[i]
            day_id.append(append_i)

    day_id.insert(0, 0)
    return day_id


class clssimp(nn.Module):
    def __init__(self, ch=2880, num_classes=20):

        super(clssimp, self).__init__()
        self.pool = nn.AdaptiveAvgPool1d(output_size=(ch))
        self.way1 = nn.Sequential(
            nn.Linear(ch, 1000, bias=True),
            nn.BatchNorm1d(1000),
            nn.ReLU(inplace=True),
        )
        self.way2 = nn.Sequential(
            nn.Linear(1000, 512, bias=True),
            nn.BatchNorm1d(512),
            nn.ReLU(inplace=True),
        )
        self.way3 = nn.Sequential(
            nn.Linear(512, 100, bias=True),
            nn.BatchNorm1d(100),
            nn.ReLU(inplace=True),
        )
        self.cls = nn.Linear(100, num_classes, bias=True)

    def forward(self, x):
        # bp()
        x = self.pool(x[None, :])
        x = x.reshape(x.size(1), -1)
        x = self.way1(x)
        x = self.way2(x)
        x = self.way3(x)
        logits = self.cls(x)
        return logits

    def intermediate_forward(self, x):
        x = self.pool(x[None, :])
        x = x.reshape(x.size(1), -1)
        x = self.way1(x)
        x = self.way2(x)
        x = self.way3(x)
        return x


class AE(nn.Module):
    def __init__(self, **kwargs):
        super().__init__()
        self.conv1 = nn.Conv1d(1, 1, 5, padding=2)
        self.bn = torch.nn.BatchNorm1d(1, eps=0.001, momentum=0.99)
        self.encoder_hidden_layer = nn.Linear(
            in_features=kwargs["input_shape"], out_features=128
        )
        self.encoder_output_layer = nn.Linear(in_features=128, out_features=128)
        self.decoder_hidden_layer = nn.Linear(in_features=128, out_features=128)
        self.decoder_output_layer = nn.Linear(
            in_features=128, out_features=kwargs["input_shape"]
        )

    def forward(self, features):
        features = self.conv1(features.reshape(features.shape[0], 1, features.shape[1]))
        features = torch.relu(self.bn(features)).reshape(
            features.shape[0], features.shape[2]
        )
        activation = self.encoder_hidden_layer(features)
        activation = self.encoder_hidden_layer(features)
        activation = torch.relu(activation)
        code = self.encoder_output_layer(activation)
        code = torch.relu(code)
        activation = self.decoder_hidden_layer(code)
        activation = torch.relu(activation)
        reconstructed = self.decoder_output_layer(activation)
        # reconstructed = torch.relu(reconstructed)
        return code, reconstructed



class AutoSort:
    def __init__(self,  ch_num, samplepoints, 
    loc_dim, device, set_shank_id,save_dir, pos_weight_noise=None,pos_weight_label=None):
        self.clsfier_noise = clssimp((ch_num+1)*samplepoints+loc_dim , 2).to(device)
        self.clsfier_label = clssimp((ch_num+1)*samplepoints+loc_dim , len(set_shank_id)).to(device)

        self.optimizer = torch.optim.Adam([
                                {'params': self.clsfier_noise.parameters()},
                                {'params': self.clsfier_label.parameters()},
                                ], lr=1e-4)


        self.criterion = nn.MSELoss()
        self.bceloss = nn.BCEWithLogitsLoss(pos_weight=pos_weight_noise)
        self.bceloss_label = nn.BCEWithLogitsLoss(pos_weight=pos_weight_label)
        

        self.save_model_path_1 = save_dir+ 'multitask_single_wave_noise_ae.pth'
        self.save_model_path_2 = save_dir+ 'multitask_single_wave_clsfier_noise_clsfier.pth'
        self.save_model_path_3 = save_dir+ 'multitask_single_wave_clsfier_label_clsfier.pth'

        self.set_shank_id = set_shank_id

    def save_model(self):
        torch.save(self.clsfier_noise.state_dict(), self.save_model_path_2)
        torch.save(self.clsfier_label.state_dict(), self.save_model_path_3)

    def load_model(self):
        # self.model.load_state_dict(torch.load(save_model_path_1))
        self.clsfier_noise.load_state_dict(torch.load(self.save_model_path_2))
        self.clsfier_label.load_state_dict(torch.load(self.save_model_path_3))

    def to_device(self, device):
        self.clsfier_noise.to(device)
        self.clsfier_label.to(device)
        self.bceloss.to(device)
        self.bceloss_label.to(device)

    def train(self):
        self.clsfier_noise.train()
        self.clsfier_label.train()
        
    def eval(self):
        self.clsfier_noise.eval()
        self.clsfier_label.eval()      

    def iter_model(self, batch_features, classify_labels, labels,
        single_waveform,pred_loc):
        self.optimizer.zero_grad()

        codes=batch_features
        codes = torch.cat((codes, single_waveform), axis=1)
        codes = torch.cat((codes, pred_loc), axis=1)

        cls_output = self.clsfier_noise(codes.float())


        test = labels[:,1]==1
        if sum(test)>1:
            cls_label_output = self.clsfier_label(codes.float()[test,:])
            train_loss3 = 1000*  self.bceloss_label(cls_label_output, classify_labels[test,:len(self.set_shank_id)])
        else:
            train_loss3 = torch.tensor(0)


        train_loss1 = 0 
        train_loss2 = 1000*  self.bceloss(cls_output, labels)


        train_loss = train_loss1 + train_loss2 + train_loss3
        train_loss.backward()
        self.optimizer.step()
        return train_loss1, train_loss2.item(), train_loss3.item(), test


    def iter_model_eval(self, batch_features, classify_labels, labels,
        single_waveform,pred_loc):

        codes=batch_features
        codes = torch.cat((codes, single_waveform), axis=1)
        codes = torch.cat((codes, pred_loc), axis=1)

        cls_output = self.clsfier_noise(codes.float())
        gt = torch.argmax(labels, axis=1)
        pred = torch.argmax(cls_output, axis=1)

        test = labels[:,1]==1
        if sum(test)>1:
            cls_label_output = self.clsfier_label(codes.float()[test,:])
            pred_class = torch.argmax(cls_label_output,axis=1)
            gt_label_class = torch.argmax(classify_labels[test, :len(self.set_shank_id)], axis=1)
            train_loss3 = 1000*  self.bceloss_label(cls_label_output, classify_labels[test,:len(self.set_shank_id)])
        else:
            train_loss3 = torch.tensor(0)
            gt_label_class=torch.tensor([])
            pred_class=torch.tensor([])

        train_loss1 = 0 
        train_loss2 = 1000*  self.bceloss(cls_output, labels)
        train_loss = train_loss1 + train_loss2 + train_loss3
        return train_loss1, train_loss2.item(), train_loss3.item(), gt, pred, gt_label_class, pred_class


    def iter_model_eval_umap(self, batch_features, classify_labels, labels,
        single_waveform,pred_loc):

        codes=batch_features
        codes = torch.cat((codes, single_waveform), axis=1)
        codes = torch.cat((codes, pred_loc), axis=1)

        cls_output = self.clsfier_noise(codes.float())
        codestest = self.clsfier_noise.intermediate_forward(codes.float())

        gt = torch.argmax(labels, axis=1)
        pred = torch.argmax(cls_output, axis=1)

        # test = cls_output[:,1]==1
        # if sum(test)>1:
        cls_label_output = self.clsfier_label(codes.float())
        codestest_label = self.clsfier_label.intermediate_forward(codes.float())

        pred_class = torch.argmax(cls_label_output,axis=1)
        gt_label_class = torch.argmax(classify_labels[:, :len(self.set_shank_id)], axis=1)

        prob = nnf.softmax(cls_label_output, dim=1)

        return gt, pred, gt_label_class, pred_class, codestest, codestest_label, prob



import numpy as np
import torch
from torch.utils import data
import pickle


def location_cal(sensor_positions, batch_features):
    NumChannels = batch_features.shape[1]
    location_day = []

    b_max = batch_features.max(-1)
    b_min = batch_features.min(-1)
    amplitudes = b_max-b_min
    # amplitudes_multi = np.multiply(amplitudes,amplitudes)
    # amplitudes = np.multiply(amplitudes_multi,amplitudes)
    amplitudes =np.square(amplitudes)
    amplitudes = np.square(amplitudes)
    sum_square_amplitute=np.sum(amplitudes,axis=1)

    location_day=[]
    for ij in range(sensor_positions.shape[1]):
        x=np.dot(sensor_positions[:, ij] , amplitudes.T)
        x=np.divide(x, sum_square_amplitute)
        location_day.append(x)
    # y=np.dot(sensor_positions[:, 1] , amplitudes.T)
    # y=np.divide(y, sum_square_amplitute)
    #
    # # location_day = [x, y]
    # z=np.dot(sensor_positions[:, 2] , amplitudes.T)
    # z=np.divide(z, sum_square_amplitute)
    # location_day=[x,y,z]
    location_day=np.array(location_day).T
    return location_day


def location_cal_group(sensor_positions, batch_features,group_id):
    group_batch = sensor_positions[:,-1]
    location_day=np.zeros((batch_features.shape[0],3))
    for i in np.unique(group_batch):
        care_loc = np.where(group_batch==i)[0]
        look_spike_loc = np.nonzero(np.in1d(group_id, care_loc))[0]
        location_day_batch = location_cal(sensor_positions[care_loc,:], batch_features[look_spike_loc,:,:][:,care_loc,:])
        location_day[look_spike_loc,:] = location_day_batch
    return location_day



class waveformLoader(data.Dataset):
    def __init__(self, root, shank_channel ,
                 sensor_positions,Keep_id=None):
        with (open(root + "X_waveform.pkl", "rb")) as openfile:
            datafile = pickle.load(openfile)
        try:
            with (open(root + "Y_spike_id.pkl", "rb")) as openfile:
                GT = pickle.load(openfile)
        except FileNotFoundError:
            GT = np.zeros(datafile.shape[0])-1
        with (open(root + "Y_spike_id_noise.pkl", "rb")) as openfile:
            channel_id = np.array(pickle.load(openfile))

        if Keep_id is None:
            Keep_id = np.unique(GT)
            Keep_id = list(Keep_id[Keep_id != -1])
            self.keep_id = Keep_id

        mask = ~np.isin(GT, Keep_id)
        GT = np.array(GT)

        GT_binary = np.zeros((GT.shape[0], 2))
        GT_binary[list(mask), 0] = 1
        GT_binary[~mask, 1] = 1

        self.GT_unique = Keep_id + [-1]
        self.GT_binary = GT_binary
        
        self.Img_single = datafile[np.arange(datafile.shape[0]), np.array(channel_id).astype('int'), :]

        self.GT_LIST = GT

        GT_array = np.zeros((len(GT), len(Keep_id)))
        for idx, unique_id in enumerate(Keep_id):
            rmv_list = np.where(np.array(GT) == unique_id)[0]
            GT_array[rmv_list, idx] = 1
        self.GT = GT_array

        self.Img = datafile

        self.pos_weight_noise =torch.tensor( [-np.sum(self.GT_binary[:,0]-1)/np.sum(self.GT_binary[:,0]),
                                 -np.sum(self.GT_binary[:,1]-1)/np.sum(self.GT_binary[:,1])])
        self.pos_weight_label = torch.tensor([-(np.sum(self.GT[:,i]-1)+sum(np.sum(GT_array,axis=1)==0))/np.sum(self.GT[:,i]) for i in range(self.GT.shape[1])])

        pred_location = location_cal_group(sensor_positions, datafile, channel_id)

        self.pred_location = pred_location
        print('pred_location',pred_location.shape)
        self.n_classes = len(set(self.GT_unique))

    def __len__(self):
        return len(self.GT)

    def __getitem__(self, index):
        return self.Img[index, ...] ,  self.GT[index, ...], self.GT_binary[index, ...], self.Img_single[index, ...],self.pred_location[index,...]




import argparse
import numpy as np


parser = argparse.ArgumentParser(description='AutoSort')

parser.add_argument('--day_id_str', type=list, default=['0411','0416'], help='day id every day.')
parser.add_argument('--group', type=np.array, default=np.arange(30), help='keep channels')
parser.add_argument('--samplepoints', type=int, default=30, help='samplepoints')
parser.add_argument('--cluster_path', type=str, default='./AutoSort_data/',
                    help='cluster path.')
parser.add_argument('--seed_all', type=int, default=0, help='seed')
parser.add_argument('--set_time', type=int, default=0, help='set_time')
parser.add_argument('--test_time', type=list, default=[1], help='test_time')

args = parser.parse_args()





#! /bin/env python
#
# Michael Gibson 23 April 2015


import sys, struct, os


def read_qstring(fid):
    """Read Qt style QString.

    The first 32-bit unsigned number indicates the length of the string (in bytes).
    If this number equals 0xFFFFFFFF, the string is null.

    Strings are stored as unicode.
    """

    (length,) = struct.unpack("<I", fid.read(4))
    if length == int("ffffffff", 16):
        return ""

    if length > (os.fstat(fid.fileno()).st_size - fid.tell() + 1):
        print(length)
        raise Exception("Length too long.")

    # convert length from bytes to 16-bit Unicode words
    length = int(length / 2)

    data = []
    for i in range(0, length):
        (c,) = struct.unpack("<H", fid.read(2))
        data.append(c)

    if sys.version_info >= (3, 0):
        a = "".join([chr(c) for c in data])
    else:
        a = "".join([unichr(c) for c in data])

    return a


if __name__ == "__main__":
    a = read_qstring(open(sys.argv[1], "rb"))
    print(a)



#! /bin/env python
#
# Michael Gibson 27 April 2015
# Modified Adrian Foy Sep 2018


def data_to_result(header, data, data_present):
    """Moves the header and data (if present) into a common object."""

    result = {}
    if header["num_amplifier_channels"] > 0 and data_present:
        result["t_amplifier"] = data["t_amplifier"]
    if header["num_aux_input_channels"] > 0 and data_present:
        result["t_aux_input"] = data["t_aux_input"]
    if header["num_supply_voltage_channels"] > 0 and data_present:
        result["t_supply_voltage"] = data["t_supply_voltage"]
    if header["num_board_adc_channels"] > 0 and data_present:
        result["t_board_adc"] = data["t_board_adc"]
    if (
        header["num_board_dig_in_channels"] > 0
        or header["num_board_dig_out_channels"] > 0
    ) and data_present:
        result["t_dig"] = data["t_dig"]
    if header["num_temp_sensor_channels"] > 0 and data_present:
        result["t_temp_sensor"] = data["t_temp_sensor"]

    if header["num_amplifier_channels"] > 0:
        result["spike_triggers"] = header["spike_triggers"]

    result["notes"] = header["notes"]
    result["frequency_parameters"] = header["frequency_parameters"]

    if header["version"]["major"] > 1:
        result["reference_channel"] = header["reference_channel"]

    if header["num_amplifier_channels"] > 0:
        result["amplifier_channels"] = header["amplifier_channels"]
        if data_present:
            result["amplifier_data"] = data["amplifier_data"]

    if header["num_aux_input_channels"] > 0:
        result["aux_input_channels"] = header["aux_input_channels"]
        if data_present:
            result["aux_input_data"] = data["aux_input_data"]

    if header["num_supply_voltage_channels"] > 0:
        result["supply_voltage_channels"] = header["supply_voltage_channels"]
        if data_present:
            result["supply_voltage_data"] = data["supply_voltage_data"]

    if header["num_board_adc_channels"] > 0:
        result["board_adc_channels"] = header["board_adc_channels"]
        if data_present:
            result["board_adc_data"] = data["board_adc_data"]

    if header["num_board_dig_in_channels"] > 0:
        result["board_dig_in_channels"] = header["board_dig_in_channels"]
        if data_present:
            result["board_dig_in_data"] = data["board_dig_in_data"]

    if header["num_board_dig_out_channels"] > 0:
        result["board_dig_out_channels"] = header["board_dig_out_channels"]
        if data_present:
            result["board_dig_out_data"] = data["board_dig_out_data"]

    return result



from autosort_neuron.model import *
from autosort_neuron.waveform_loader import *
from autosort_neuron.utils import *
from autosort_neuron.sma import *
from autosort_neuron.sorting import *
from autosort_neuron.detection import *
from autosort_neuron.config import *
from autosort_neuron.run import *



import sys, struct, math, os, time

sys.path.append("./autosort_neuron/")
import numpy as np
from tqdm import tqdm
from intanutil.read_header import read_header
from intanutil.get_bytes_per_data_block import get_bytes_per_data_block
from intanutil.read_one_data_block import read_one_data_block
from intanutil.notch_filter import notch_filter
from intanutil.data_to_result import data_to_result
from scipy import signal
import os
import spikeinterface

spikeinterface.__version__
import spikeinterface
import spikeinterface.extractors as se
import spikeinterface.sorters as ss
# import spikeinterface.toolkit as st
import spikeinterface.widgets as sw
import matplotlib.pyplot as plt
import numpy as np
import shutil
import pandas as pd
from spikeinterface.core.npzsortingextractor import NpzSortingExtractor
from pylab import *
import pylab
from probeinterface import Probe
from probeinterface.plotting import plot_probe
import sys
import warnings

warnings.filterwarnings("ignore")


def sorting_unit_show(we, recording_cmr, sorting, pack_folder, waveform_folder):
    # plot_probe(mesh_probe,with_channel_index=True)

    fig, axs = plt.subplots(
        int(np.ceil(len(sorting.unit_ids) / 4)),
        4,
        figsize=(20, 5 * np.ceil(len(sorting.unit_ids) / 4)),
    )
    sw.plot_unit_templates(we, unit_ids=sorting.unit_ids, axes=axs)
    plt.savefig(waveform_folder + "/templates_map.png", dpi=300)

    extremum_channels_ids = spikeinterface.core.get_template_extremum_channel(we, peak_sign="neg")

    colors = []
    cm = get_cmap("rainbow")
    NUM_COLORS = len(sorting.unit_ids)
    for i in range(NUM_COLORS):
        colors.append(cm(1.0 * i / NUM_COLORS))  # color will now be an RGBA tuple

    fig, axs = plt.subplots(
        int(np.ceil(len(sorting.unit_ids) / 4)),
        4,
        figsize=(20, 5 * np.ceil(len(sorting.unit_ids) / 4)),
    )

    for i, unit_id in enumerate(sorting.unit_ids):
        template = we.get_template(unit_id)
        ax = axs[int(np.floor(i / 4)), int(np.mod(i, 4))]
        ax.plot(
            template[:, extremum_channels_ids[unit_id]].T,
            lw=3,
            label=unit_id,
            color=colors[i],
        )
        ax.set_title(f"template{unit_id}")

    plt.savefig(waveform_folder + "/extremum_templates_map.png", dpi=300)


def read_data(filename):
    """Reads Intan Technologies RHD2000 data file generated by evaluation board GUI.

    Data are returned in a dictionary, for future extensibility.
    """

    tic = time.time()
    fid = open(filename, "rb")
    filesize = os.path.getsize(filename)

    header = read_header(fid)

    print(
        "Found {} amplifier channel{}.".format(
            header["num_amplifier_channels"], plural(header["num_amplifier_channels"])
        )
    )
    print(
        "Found {} auxiliary input channel{}.".format(
            header["num_aux_input_channels"], plural(header["num_aux_input_channels"])
        )
    )
    print(
        "Found {} supply voltage channel{}.".format(
            header["num_supply_voltage_channels"],
            plural(header["num_supply_voltage_channels"]),
        )
    )
    print(
        "Found {} board ADC channel{}.".format(
            header["num_board_adc_channels"], plural(header["num_board_adc_channels"])
        )
    )
    print(
        "Found {} board digital input channel{}.".format(
            header["num_board_dig_in_channels"],
            plural(header["num_board_dig_in_channels"]),
        )
    )
    print(
        "Found {} board digital output channel{}.".format(
            header["num_board_dig_out_channels"],
            plural(header["num_board_dig_out_channels"]),
        )
    )
    print(
        "Found {} temperature sensors channel{}.".format(
            header["num_temp_sensor_channels"],
            plural(header["num_temp_sensor_channels"]),
        )
    )
    print("")

    # Determine how many samples the data file contains.
    bytes_per_block = get_bytes_per_data_block(header)

    # How many data blocks remain in this file?
    data_present = False
    bytes_remaining = filesize - fid.tell()
    if bytes_remaining > 0:
        data_present = True

    if bytes_remaining % bytes_per_block != 0:
        raise Exception(
            "Something is wrong with file size : should have a whole number of data blocks"
        )

    num_data_blocks = int(bytes_remaining / bytes_per_block)

    num_amplifier_samples = header["num_samples_per_data_block"] * num_data_blocks
    num_aux_input_samples = int(
        (header["num_samples_per_data_block"] / 4) * num_data_blocks
    )
    num_supply_voltage_samples = 1 * num_data_blocks
    num_board_adc_samples = header["num_samples_per_data_block"] * num_data_blocks
    num_board_dig_in_samples = header["num_samples_per_data_block"] * num_data_blocks
    num_board_dig_out_samples = header["num_samples_per_data_block"] * num_data_blocks

    record_time = num_amplifier_samples / header["sample_rate"]

    if data_present:
        print(
            "File contains {:0.3f} seconds of data.  Amplifiers were sampled at {:0.2f} kS/s.".format(
                record_time, header["sample_rate"] / 1000
            )
        )
    else:
        print(
            "Header file contains no data.  Amplifiers were sampled at {:0.2f} kS/s.".format(
                header["sample_rate"] / 1000
            )
        )

    if data_present:
        # Pre-allocate memory for data.
        print("")
        print("Allocating memory for data...")

        data = {}
        if (header["version"]["major"] == 1 and header["version"]["minor"] >= 2) or (
            header["version"]["major"] > 1
        ):
            data["t_amplifier"] = np.zeros(num_amplifier_samples, dtype=np.int)
        else:
            data["t_amplifier"] = np.zeros(num_amplifier_samples, dtype=np.uint)

        data["amplifier_data"] = np.zeros(
            [header["num_amplifier_channels"], num_amplifier_samples], dtype=np.uint
        )
        data["aux_input_data"] = np.zeros(
            [header["num_aux_input_channels"], num_aux_input_samples], dtype=np.uint
        )
        data["supply_voltage_data"] = np.zeros(
            [header["num_supply_voltage_channels"], num_supply_voltage_samples],
            dtype=np.uint,
        )
        data["temp_sensor_data"] = np.zeros(
            [header["num_temp_sensor_channels"], num_supply_voltage_samples],
            dtype=np.uint,
        )
        data["board_adc_data"] = np.zeros(
            [header["num_board_adc_channels"], num_board_adc_samples], dtype=np.uint
        )

        # by default, this script interprets digital events (digital inputs and outputs) as booleans
        # if unsigned int values are preferred(0 for False, 1 for True), replace the 'dtype=np.bool' argument with 'dtype=np.uint' as shown
        # the commented line below illustrates this for digital input data; the same can be done for digital out

        # data['board_dig_in_data'] = np.zeros([header['num_board_dig_in_channels'], num_board_dig_in_samples], dtype=np.uint)
        data["board_dig_in_data"] = np.zeros(
            [header["num_board_dig_in_channels"], num_board_dig_in_samples],
            dtype=np.bool,
        )
        data["board_dig_in_raw"] = np.zeros(num_board_dig_in_samples, dtype=np.uint)

        data["board_dig_out_data"] = np.zeros(
            [header["num_board_dig_out_channels"], num_board_dig_out_samples],
            dtype=np.bool,
        )
        data["board_dig_out_raw"] = np.zeros(num_board_dig_out_samples, dtype=np.uint)

        # Read sampled data from file.
        print("Reading data from file...")

        # Initialize indices used in looping
        indices = {}
        indices["amplifier"] = 0
        indices["aux_input"] = 0
        indices["supply_voltage"] = 0
        indices["board_adc"] = 0
        indices["board_dig_in"] = 0
        indices["board_dig_out"] = 0

        print_increment = 10
        percent_done = print_increment
        for i in range(num_data_blocks):
            read_one_data_block(data, header, indices, fid)

            # Increment indices
            indices["amplifier"] += header["num_samples_per_data_block"]
            indices["aux_input"] += int(header["num_samples_per_data_block"] / 4)
            indices["supply_voltage"] += 1
            indices["board_adc"] += header["num_samples_per_data_block"]
            indices["board_dig_in"] += header["num_samples_per_data_block"]
            indices["board_dig_out"] += header["num_samples_per_data_block"]

            fraction_done = 100 * (1.0 * i / num_data_blocks)
            if fraction_done >= percent_done:
                print("{}% done...".format(percent_done))
                percent_done = percent_done + print_increment

        # Make sure we have read exactly the right amount of data.
        bytes_remaining = filesize - fid.tell()
        if bytes_remaining != 0:
            raise Exception("Error: End of file not reached.")

    # Close data file.
    fid.close()

    if data_present:
        print("Parsing data...")

        # Extract digital input channels to separate variables.
        for i in range(header["num_board_dig_in_channels"]):
            data["board_dig_in_data"][i, :] = np.not_equal(
                np.bitwise_and(
                    data["board_dig_in_raw"],
                    (1 << header["board_dig_in_channels"][i]["native_order"]),
                ),
                0,
            )

        # Extract digital output channels to separate variables.
        for i in range(header["num_board_dig_out_channels"]):
            data["board_dig_out_data"][i, :] = np.not_equal(
                np.bitwise_and(
                    data["board_dig_out_raw"],
                    (1 << header["board_dig_out_channels"][i]["native_order"]),
                ),
                0,
            )

        # Scale voltage levels appropriately.
        data["amplifier_data"] = np.multiply(
            0.195, (data["amplifier_data"].astype(np.int32) - 32768)
        )  # units = microvolts
        data["aux_input_data"] = np.multiply(
            37.4e-6, data["aux_input_data"]
        )  # units = volts
        data["supply_voltage_data"] = np.multiply(
            74.8e-6, data["supply_voltage_data"]
        )  # units = volts
        if header["eval_board_mode"] == 1:
            data["board_adc_data"] = np.multiply(
                152.59e-6, (data["board_adc_data"].astype(np.int32) - 32768)
            )  # units = volts
        elif header["eval_board_mode"] == 13:
            data["board_adc_data"] = np.multiply(
                312.5e-6, (data["board_adc_data"].astype(np.int32) - 32768)
            )  # units = volts
        else:
            data["board_adc_data"] = np.multiply(
                50.354e-6, data["board_adc_data"]
            )  # units = volts
        data["temp_sensor_data"] = np.multiply(
            0.01, data["temp_sensor_data"]
        )  # units = deg C

        # Check for gaps in timestamps.
        num_gaps = np.sum(
            np.not_equal(data["t_amplifier"][1:] - data["t_amplifier"][:-1], 1)
        )
        if num_gaps == 0:
            print("No missing timestamps in data.")
        else:
            print(
                "Warning: {0} gaps in timestamp data found.  Time scale will not be uniform!".format(
                    num_gaps
                )
            )

        # Scale time steps (units = seconds).
        data["t_amplifier"] = data["t_amplifier"] / header["sample_rate"]
        data["t_aux_input"] = data["t_amplifier"][range(0, len(data["t_amplifier"]), 4)]
        data["t_supply_voltage"] = data["t_amplifier"][
            range(0, len(data["t_amplifier"]), header["num_samples_per_data_block"])
        ]
        data["t_board_adc"] = data["t_amplifier"]
        data["t_dig"] = data["t_amplifier"]
        data["t_temp_sensor"] = data["t_supply_voltage"]

        # If the software notch filter was selected during the recording, apply the
        # same notch filter to amplifier data here.
        if header["notch_filter_frequency"] > 0 and header["version"]["major"] < 3:
            print("Applying notch filter...")

            print_increment = 10
            percent_done = print_increment
            for i in range(header["num_amplifier_channels"]):
                data["amplifier_data"][i, :] = notch_filter(
                    data["amplifier_data"][i, :],
                    header["sample_rate"],
                    header["notch_filter_frequency"],
                    10,
                )

                fraction_done = 100 * (i / header["num_amplifier_channels"])
                if fraction_done >= percent_done:
                    print("{}% done...".format(percent_done))
                    percent_done += print_increment
    else:
        data = []

    # Move variables to result struct.
    result = data_to_result(header, data, data_present)

    print("Done!  Elapsed time: {0:0.1f} seconds".format(time.time() - tic))
    return result


def plural(n):
    """Utility function to optionally pluralize words based on the value of n."""

    if n == 1:
        return ""
    else:
        return "s"


def sorting_day_split(
    sorting, date_id_all, day_length, pack_folder, sorting_save_name="firings_inlier"
):
    sampling_freq = sorting.get_sampling_frequency()

    fig, ax = plt.subplots(1, 1, figsize=(20, 40))
    sw.plot_rasters(sorting, time_range=(0, np.sum(day_length)), ax=ax)

    colors = []
    cm = pylab.get_cmap("rainbow")
    NUM_COLORS = len(day_length)
    for i in range(NUM_COLORS):
        colors.append(cm(1.0 * i / NUM_COLORS))  # color will now be an RGBA tuple

    for i in range(len(day_length)):
        ax.axvspan(
            np.sum(day_length[:i]) / sampling_freq,
            np.sum(day_length[: (i + 1)]) / sampling_freq,
            facecolor=colors[i],
            alpha=0.1,
        )
    plt.show()
    plt.savefig(pack_folder + "/sorting/" + sorting_save_name + ".png", dpi=300)

    for i in range(len(day_length)):
        pack_folder_i = pack_folder + "/" + date_id_all[i] + "/"

        if os.path.exists(pack_folder_i) == False:
            os.mkdir(pack_folder_i)

        start_frame = np.sum(day_length[:i])
        end_frame = np.sum(day_length[: (i + 1)])

        sub_sorting = sorting.frame_slice(start_frame, end_frame)

        keep_unit_ids = []
        for unit_id in sub_sorting.unit_ids:
            spike_train = sub_sorting.get_unit_spike_train(unit_id=unit_id)
            n = spike_train.size
            if n > 20:
                keep_unit_ids.append(unit_id)

        curated_sub_sorting = sub_sorting.select_units(
            unit_ids=keep_unit_ids, renamed_unit_ids=None
        )

        save_path = pack_folder_i + "/sorting/" + sorting_save_name + ".npz"
        if os.path.exists(pack_folder_i + "/sorting/") == False:
            os.mkdir(pack_folder_i + "/sorting/")
        NpzSortingExtractor.write_sorting(curated_sub_sorting, save_path)


def get_sorting_info(
    data_folder_all,
    pack_folder,
    waveform_save_folder,
    input_state="_merged",
    we_load_if_exists=True,
    freq_max=3000,
    freq_min=300,
):
    recording = spikeinterface.core.base.BaseExtractor.load_from_folder(data_folder_all)
    recording_f = spikeinterface.preprocessing.bandpass_filter(recording, freq_min=freq_min, freq_max=freq_max)
    recording_cmr = spikeinterface.preprocessing.common_reference(
        recording_f, reference="global", operator="average"
    )

    sorting = se.NpzSortingExtractor(
        pack_folder + "/sorting/firings" + input_state + ".npz"
    )

    we = spikeinterface.extract_waveforms(
        recording_cmr,
        sorting,
        waveform_save_folder,
        load_if_exists=we_load_if_exists,
        overwrite=False,
        ms_before=1,
        ms_after=2,
        max_spikes_per_unit=1000000,
        n_jobs=1,
        chunk_size=30000,
    )
    return recording_cmr, sorting, we


def recording_shank_split(
    recording,
    sorting,
    we,
    pack_folder,
    input_state="inlier",
    we_load_if_exists=True,
    probe_groups=None,
):

    extremum_channels_ids = spikeinterface.core.get_template_extremum_channel(we, peak_sign="neg")

    if probe_groups is None:
        probe_groups = np.arange(0, 30)

    NUmShanks = np.shape(np.unique(probe_groups))[0]

    recording.set_property("group", probe_groups, ids=None)
    slice_recording = recording.split_by()
    slice_sorting = []
    slice_we = []
    slice_unit_ids = []

    for shank_id in np.arange(NUmShanks):
        shank_channel_ids = np.where(probe_groups == shank_id)[0]
        shank_unit_ids = sorting.unit_ids[
            np.where(
                np.isin(
                    np.array(list(extremum_channels_ids.values())), shank_channel_ids
                )
            )[0]
        ]
        print(f"Shank:{shank_id+1}, units: {shank_unit_ids}")
        slice_unit_ids.append(shank_unit_ids)

        shank_recording = slice_recording[shank_id]
        shank_sorting = sorting.select_units(
            unit_ids=shank_unit_ids, renamed_unit_ids=None
        )
        slice_sorting.append(shank_sorting)

        shank_waveform_folder = (
            pack_folder + f"/waveforms_shank{shank_id+1}_{input_state}"
        )
        if we_load_if_exists == False:
            if os.path.exists(shank_waveform_folder):
                shutil.rmtree(shank_waveform_folder)

        shank_we = spikeinterface.extract_waveforms(
            shank_recording,
            shank_sorting,
            shank_waveform_folder,
            load_if_exists=we_load_if_exists,
            ms_before=1,
            ms_after=2.0,
            max_spikes_per_unit=10000,
            n_jobs=1,
            chunk_size=300,
        )

        slice_we.append(shank_we)

    return slice_recording, slice_sorting, slice_we


def shank_waveform_show(slice_sorting, slice_we, waveform_show=False, save_folder=None):
    NumShanks = len(slice_sorting)
    for shank_id in np.arange(NumShanks):
        shank_sorting = slice_sorting[shank_id]
        shank_we = slice_we[shank_id]

        # waveform
        if shank_sorting is not None:
            if len(shank_sorting.unit_ids) > 0:
                fig, axs = plt.subplots(
                    int(np.ceil(len(shank_sorting.unit_ids) / 8)),
                    8,
                    figsize=(20, 3 * np.ceil(len(shank_sorting.unit_ids) / 8)),
                )
                sw.plot_unit_templates(
                    shank_we, unit_ids=shank_sorting.unit_ids, axes=axs
                )
                fig.suptitle(f"Shank{shank_id+1}: templates")
                fig.tight_layout()
                if save_folder is not None:
                    plt.savefig(
                        save_folder + f"/template_map_shank{shank_id+1}.png", dpi=300
                    )

                if waveform_show == True:
                    fig, axs = plt.subplots(
                        int(np.ceil(len(shank_sorting.unit_ids) / 8)),
                        8,
                        figsize=(20, 3 * np.ceil(len(shank_sorting.unit_ids) / 8)),
                    )
                    sw.plot_unit_waveforms(
                        shank_we, unit_ids=shank_sorting.unit_ids, axes=axs
                    )
                    fig.suptitle(f"Shank{shank_id+1}: waveforms")
                    fig.tight_layout()
                    if save_folder is not None:
                        plt.savefig(
                            save_folder
                            + f"/extremum_waveforms_map_shank{shank_id+1}.png",
                            dpi=300,
                        )


def shank_ISI_show(slice_sorting, slice_we, save_folder=None):
    NumShanks = len(slice_sorting)

    unit_ids = [list(slice_sorting[shank_id].unit_ids) for shank_id in range(NumShanks)]
    NumUnits = np.sum([len(unit_ids[shank_id]) for shank_id in range(NumShanks)])
    Max_NumUnits = np.max([len(unit_ids[shank_id]) for shank_id in range(NumShanks)])

    # Set colors
    cm = pylab.get_cmap("rainbow")
    colors = []
    for shank_id in np.arange(NumShanks):
        unit_shank_ids = unit_ids[shank_id]
        colors_shank = []
        for j in range(len(unit_shank_ids)):
            idx = j * NumShanks + shank_id
            colors_shank.append(cm(1.0 * idx / (Max_NumUnits * NumShanks)))
        colors.append(colors_shank)

    # Plot ISI histograms
    for shank_id in np.arange(NumShanks):
        shank_sorting = slice_sorting[shank_id]
        shank_we = slice_we[shank_id]
        unit_shank_ids = unit_ids[shank_id]
        if int(np.ceil(len(unit_shank_ids) / 8)) > 0:
            fig, axs = plt.subplots(
                int(np.ceil(len(unit_shank_ids) / 8)),
                8,
                figsize=(20, 2.5 * np.ceil(len(unit_shank_ids) / 8)),
            )

            for i, unit_id in enumerate(unit_shank_ids):
                if int(np.ceil(len(unit_shank_ids) / 8)) == 1:
                    ax = axs[i]
                else:
                    ax = axs[int(np.floor(i / 8)), int(np.mod(i, 8))]

                ax.set_title(f"unit{unit_id}")
                sw.plot_isi_distribution(
                    shank_sorting, unit_ids=np.array(unit_id, ndmin=1), axes=ax
                )

            fig.suptitle(f"Shank{shank_id+1}: ISI Histograms")
            fig.tight_layout()

        if save_folder is not None:
            plt.savefig(save_folder + f"/ISI_histograms_shank{shank_id+1}.png", dpi=300)


def metrics_cal(slice_we, isi_threshold_ms=1.5):
    NumShanks = len(slice_we)

    metrics_all = []
    slice_curated_ids = [None] * NumShanks
    slice_curated_sorting = [None] * NumShanks
    slice_curated_we = [None] * NumShanks

    for shank_id in np.arange(NumShanks):
        shank_we = slice_we[shank_id]
        if shank_we is not None:

            pc = spikeinterface.postprocessing.compute_principal_components(
                shank_we, load_if_exists=True, n_components=3, mode="by_channel_local"
            )
            metrics = spikeinterface.qualitymetrics.compute_quality_metrics(
                shank_we,
                metric_names=[
                    "snr",
                    "isi_violation",
                    "nearest_neighbor",
                    "isolation_distance",
                    "l_ratio",
                    "firing_rate",
                ],
            )
            isi_violations_rate = compute_isi_violations(
                shank_we, isi_threshold_ms=isi_threshold_ms
            )

            metrics["shank"] = shank_id + 1
            metrics["unit_id"] = shank_we.sorting.unit_ids
            metrics["isi_violation"] = list(isi_violations_rate.values())
            metrics_all.append(metrics)

    metrics_all = pd.concat(metrics_all)

    return metrics_all


def compute_isi_violations(
    waveform_extractor, isi_threshold_ms=1.5, min_isi_ms=0, **kwargs
):
    recording = waveform_extractor.recording
    sorting = waveform_extractor.sorting
    unit_ids = sorting.unit_ids
    num_segs = sorting.get_num_segments()
    fs = recording.get_sampling_frequency()

    seg_durations = [recording.get_num_samples(i) / fs for i in range(num_segs)]
    total_duration = np.sum(seg_durations)

    isi_threshold_s = isi_threshold_ms / 1000
    min_isi_s = min_isi_ms / 1000
    isi_threshold_samples = int(isi_threshold_s * fs)

    isi_violations_rate = {}
    isi_violations_count = {}
    isi_violations_ratio = {}

    # all units converted to seconds
    for unit_id in unit_ids:
        num_violations = 0
        num_spikes = 0
        for segment_index in range(num_segs):
            spike_train = sorting.get_unit_spike_train(
                unit_id=unit_id, segment_index=segment_index
            )
            isis = np.diff(spike_train)
            num_spikes += len(spike_train)
            num_violations += np.sum(isis < isi_threshold_samples)

        isi_violations_rate[unit_id] = num_violations / num_spikes
    return isi_violations_rate


def sorting_shank_split(
    pack_folder=None,
    waveform_folder=None,
    waveform_save_folder=None,
    merge_unit_ids_pack=None,
    we_load_if_exists=False,
    output_state="merged",
    waveform_show=False,
    isi_threshold_ms=1.5,
):
    """After the sorting is done, this plots the template map, waveforms, templates, and autocorrelograms.
    Calculates metrics.
    """
    method_save_path = pack_folder
    curation_save_folder = method_save_path + f"/curation_result_{output_state}/"
    if os.path.exists(curation_save_folder) == False:
        os.mkdir(curation_save_folder)

    recording_cmr, sorting, we = get_sorting_info(
        data_folder_all,
        pack_folder,
        waveform_save_folder,
        input_state="_merged",
        we_load_if_exists=we_load_if_exists,
        freq_max=3000,
        freq_min=300,
    )

    # If applicable, merge units
    if merge_unit_ids_pack is not None:
        merged_we, merged_sorting = units_merge(
            recording_cmr,
            sorting,
            merge_unit_ids_pack,
            method_save_path,
            we_load_if_exists=we_load_if_exists,
        )
        sorting = merged_sorting
        we = merged_we

    # Plot template maps
    slice_recording, slice_sorting, slice_we = recording_shank_split(
        recording_cmr,
        sorting,
        we,
        method_save_path,
        input_state=output_state,
        we_load_if_exists=we_load_if_exists,
    )
    fig, axs = plt.subplots(
        int(np.ceil(len(sorting.unit_ids) / 5)),
        5,
        figsize=(20, 12.5 * np.ceil(len(sorting.unit_ids) / 5)),
    )
    sw.plot_unit_templates(we, unit_ids=sorting.unit_ids, axes=axs)
    plt.savefig(curation_save_folder + "/template_map.png", dpi=100)

    colors = []
    cm = pylab.get_cmap("rainbow")
    NUM_COLORS = len(sorting.unit_ids)
    for i in range(NUM_COLORS):
        colors.append(cm(1.0 * i / NUM_COLORS))  # color will now be an RGBA tuple

    # Plot extremum waveforms
    if waveform_show == True:
        extremum_channels_ids = spikeinterface.core.get_template_extremum_channel(we, peak_sign="neg")
        fig, axs = plt.subplots(
            int(np.ceil(len(sorting.unit_ids) / 5)),
            5,
            figsize=(20, 4 * np.ceil(len(sorting.unit_ids) / 5)),
        )
        for i, unit_id in enumerate(sorting.unit_ids):
            template = we.get_waveforms(unit_id)
            ax = axs[int(np.floor(i / 5)), int(np.mod(i, 5))]
            ax.plot(
                template[:, :, extremum_channels_ids[unit_id]].T,
                lw=0.3,
                label=unit_id,
                color=colors[i],
            )
            ax.set_title(f"template{unit_id}")
        plt.savefig(curation_save_folder + "/extremum_waveforms_map.png", dpi=100)

    shank_waveform_show(
        slice_sorting,
        slice_we,
        waveform_show=waveform_show,
        save_folder=curation_save_folder,
    )

    # Plot ISI histograms

    shank_ISI_show(slice_sorting, slice_we, save_folder=curation_save_folder)

    # Calculate metrics
    metrics_all = metrics_cal(slice_we, isi_threshold_ms=isi_threshold_ms)
    bad_mask = (
        (metrics_all["snr"] < 6)
        & (metrics_all["nn_hit_rate"] < 0.80)
        & (metrics_all["firing_rate"] < 0.10)
    )
    shank_qc_bad_unit_ids = bad_mask[bad_mask].index.values
    print(f"qc bad units: {shank_qc_bad_unit_ids}")
    metrics_all
    return (
        recording_cmr,
        sorting,
        we,
        slice_recording,
        slice_sorting,
        slice_we,
        metrics_all,
    )


def create_mesh_probe(positions, num_all_channels):
    mesh_probe = Probe(ndim=2, si_units="um")
    mesh_probe.set_contacts(
        positions=positions, shapes="circle", shape_params={"radius": 5}
    )

    ant = {"first_index": 0}
    mesh_probe.annotate(**ant)
    channel_indices_raw = np.arange(num_all_channels)
    channel_indices = [i for i in channel_indices_raw]
    mesh_probe.set_device_channel_indices(channel_indices)
    return mesh_probe


def stack_recordings(pack_folder_pre, mesh_probe, trigger_val=3.5, trigger=True):
    cont_data_all = []
    cont_trigger_all = []
    for dirpath, dirname, filenames in os.walk(pack_folder_pre):
        for i in sorted(filenames):
            if ".rhd" in i:
                print(dirpath + "/" + i)

                raw_data = read_data(dirpath + "/" + i)

                sampling_freq = raw_data["frequency_parameters"][
                    "amplifier_sample_rate"
                ]
                keep_list = np.arange(32)
                keep_list = np.delete(keep_list, 23)
                keep_list = np.delete(keep_list, 24)

                data = raw_data["amplifier_data"].T
                data_trigger = raw_data["board_dig_in_data"].flatten()
                if sampling_freq != 10000:
                    print("downsampling to 10000")
                    orig_sr = sampling_freq  # 10 kHz
                    new_sr = 10000  # 1 kH
                    resample_factor = orig_sr / new_sr

                    new_data = []
                    for j in tqdm(range(data.shape[1])):
                        new_data_j = signal.resample(
                            data[:, j], int(data.shape[0] / resample_factor)
                        )
                        new_data.append(new_data_j)
                    new_data = np.vstack(new_data)
                    new_data_trigger = data_trigger[:: int(resample_factor)]
                    new_data = new_data.T

                    data = new_data
                    data_trigger = new_data_trigger

                cont_data_all.append(data[:, list(keep_list)])
                if trigger:
                    cont_trigger_all.append(data_trigger)

    cont_data_all = np.vstack(cont_data_all)
    if trigger:
        cont_trigger_all = np.hstack(cont_trigger_all)

    recording = se.NumpyRecording(traces_list=cont_data_all, sampling_frequency=10000)

    recording.set_probe(mesh_probe, in_place=True)
    if trigger:
        return recording, cont_data_all, cont_trigger_all
    else:
        return recording, cont_data_all


def read_data_folder(
    data_folder_all, date_id_all, raw_data_path, mesh_probe, trigger=True
):

    recording_traces = []
    session_length_concat = []
    day_length = []
    cont_trigger_all_all = []
    if os.path.exists(data_folder_all):
        recording_concat = spikeinterface.core.base.BaseExtractor.load_from_folder(
            data_folder_all
        )
        session_length_concat = np.load(data_folder_all + "session_length.npy")
        day_length = np.load(data_folder_all + "day_length.npy")
        if trigger:
            cont_trigger_all_all = np.load(data_folder_all + "cont_trigger_all_all.npy")
    else:
        for date_id in date_id_all:
            pack_folder_pre = raw_data_path + date_id
            data_folder_pre = f"../processed_data/Ephys_{date_id}/"
            if os.path.exists(data_folder_pre):
                recording = spikeinterface.core.base.BaseExtractor.load_from_folder(
                    data_folder_pre
                )
                session_length = np.load(data_folder_pre + "session_length.npy")
                if trigger:
                    cont_trigger_all = np.load(data_folder_pre + "cont_trigger_all.npy")
            else:
                if trigger:
                    recording, session_length, cont_trigger_all = stack_recordings(
                        pack_folder_pre, mesh_probe, trigger_val=3
                    )
                else:
                    recording, session_length = stack_recordings(
                        pack_folder_pre, mesh_probe, trigger_val=3, trigger=trigger
                    )

                recording.set_probe(mesh_probe, in_place=True)

                recording = recording.save(folder=data_folder_pre)
                np.save(data_folder_pre + "session_length.npy", session_length)
                if trigger:
                    np.save(data_folder_pre + "cont_trigger_all.npy", cont_trigger_all)

            sampling_freq = recording.get_sampling_frequency()
            recording_trace = recording.get_traces()
            recording_traces.append(recording_trace)
            session_length_concat.append(session_length)
            if trigger:
                cont_trigger_all_all.append(cont_trigger_all)
            day_length.append(recording_trace.shape[0])
        recording_traces = np.vstack(recording_traces)
        session_length_concat = np.vstack(session_length_concat)
        recording_concat = se.NumpyRecording(
            traces_list=recording_traces, sampling_frequency=sampling_freq
        )
        recording_concat.set_probe(mesh_probe, in_place=True)
        recording_concat = recording_concat.save(folder=data_folder_all)
        np.save(data_folder_all + "session_length.npy", session_length_concat)
        np.save(data_folder_all + "day_length.npy", day_length)
        if trigger:
            cont_trigger_all_all = np.hstack(cont_trigger_all_all)
            np.save(data_folder_all + "cont_trigger_all_all.npy", cont_trigger_all_all)
    print(recording_concat)
    print("Num. channels = {}".format(len(recording_concat.get_channel_ids())))
    print(
        "Sampling frequency = {} Hz".format(recording_concat.get_sampling_frequency())
    )
    print("Num. timepoints seg0= {}".format(recording_concat.get_num_segments()))
    return recording_concat, day_length


def units_merge(
    recording_cmr,
    sorting,
    merge_unit_ids_pack,
    delete_unit_ids_pack,
    pack_folder,
    we_load_if_exists=True,
):
    S = sorting._sorting_segments[0]
    merged_sorting = sorting
    remove_ids = []

    for idx in range(len(merge_unit_ids_pack)):
        merge_unit_ids = merge_unit_ids_pack[idx]

        for unit_id_id, unit_id in enumerate(merge_unit_ids):
            S.spike_labels[S.spike_labels == unit_id] = merge_unit_ids[0]

        merged_sorting._sorting_segments[0] = S

        remove_ids.extend(merge_unit_ids[1:])

    remove_ids += delete_unit_ids_pack

    keep_ids = merged_sorting.unit_ids[~np.isin(merged_sorting.unit_ids, remove_ids)]

    merged_sorting = merged_sorting.select_units(
        unit_ids=keep_ids, renamed_unit_ids=None
    )

    waveform_save_folder = pack_folder + "/waveforms_merged"
    if we_load_if_exists == False:
        if os.path.exists(waveform_save_folder):
            shutil.rmtree(waveform_save_folder)
    save_sorting_merged_path = pack_folder + f"/sorting/firings_merged.npz"
    NpzSortingExtractor.write_sorting(merged_sorting, save_sorting_merged_path)
    merged_sorting = se.NpzSortingExtractor(save_sorting_merged_path)
    merged_we = spikeinterface.extract_waveforms(
        recording_cmr,
        merged_sorting,
        waveform_save_folder,
        load_if_exists=we_load_if_exists,
        overwrite=False,
        ms_before=1,
        ms_after=2,
        max_spikes_per_unit=1000000,
        n_jobs=1,
        chunk_size=30000,
    )
    sorting = merged_sorting
    we = merged_we
    return sorting, we



import math
import scipy
import numpy as np


def offline_smoother(yIn, kernSD, stepSize):
    causal = False

    if (kernSD == 0) or (yIn.shape[1] == 1):
        yOut = yIn
        return yOut

    # Filter half length
    # Go 3 standard deviations out
    fltHL = math.ceil(3 * kernSD / stepSize)

    # Length of flt is 2*fltHL + 1
    flt = scipy.stats.norm.pdf(
        np.arange(-fltHL * stepSize, fltHL * stepSize, stepSize), loc=0, scale=kernSD
    )

    [yDim, T] = shape(yIn)
    yOut = np.zeros((yDim, T))

    nm = scipy.signal.convolve(
        flt.reshape(1, -1), np.ones((1, T)), mode="full", method="auto"
    )

    for i in range(yDim):

        ys = np.divide(
            scipy.signal.convolve(flt.reshape(1, -1), yIn[i, :].reshape(1, -1)), nm
        )
        #     # Cut off edges so that result of convolution is same length
        #     # as original data
        yOut[i, :] = ys[0, fltHL : ys.shape[1] - fltHL + 1]
    return yOut


class FSM2:
    def __init__(self, k, d, tau0=None, Minv0=None, W0=None, learning_rate=None):
        self.k = k
        self.d = d
        if W0 is None:
            self.W = np.eye(k, d) / d
        else:
            self.W = W0
        self.M = np.eye(k)
        if Minv0 is None:
            self.Minv = np.eye(k)
        else:
            self.Minv = Minv0
        if tau0 is None:
            self.tau = 0.5
        else:
            self.tau = tau0
        if learning_rate is None:
            self.lr = lambda x: 1.0 / (2 * (x + 100) + 5)
        else:
            self.lr = lambda x: learning_rate

        self.t = 0
        self.outer_W = np.zeros(shape(self.W))
        self.outer_Minv = np.zeros(shape(self.Minv))
        self.y = []

    def fit_next(self, x):
        step = self.lr(self.t)
        #         print('step',step)

        z = np.dot(self.W, x.T)
        y = np.zeros(shape(z))
        #         print('z',z)
        #         print('y',y)

        # The loop below solves the linear system My = Wx
        y[0] = z[0] / self.M[0, 0]
        #         print('y[0]',y[0])
        for i in np.arange(1, self.k):
            y[i] = z[i]
            #             print('i',i,'y[i]',y[i])
            for j in np.arange(0, i):
                y[i] = y[i] - y[j] * self.M[i, j]
            #                 print('i',i,'j',j,'y[i]',y[i])

            y[i] = y[i] / self.M[i, i]
        #             print('i',i,'y[i]',y[i])

        y_up = np.tril(
            np.dot(y.reshape(y.shape[0], 1), y.reshape(y.shape[0], 1).T)
        )  # extract lower triangular component of y
        #         print('y_up',y_up)
        self.W = self.W + step * (
            y.reshape(y.shape[0], 1) * x.reshape(x.shape[0], 1).T - self.W
        )  # update W matrix
        #         print('W',self.W)
        self.M = self.M + step * (y_up - self.M)  # update M matrix
        #         print('M',self.M)
        self.t = self.t + 1
        #         print('t',self.t)
        return y



#! /bin/env python
#
# Michael Gibson 23 April 2015
# Modified Adrian Foy Sep 2018

import sys, struct
import numpy as np


def read_one_data_block(data, header, indices, fid):
    """Reads one 60 or 128 sample data block from fid into data, at the location indicated by indices."""

    # In version 1.2, we moved from saving timestamps as unsigned
    # integers to signed integers to accommodate negative (adjusted)
    # timestamps for pretrigger data['
    if (header["version"]["major"] == 1 and header["version"]["minor"] >= 2) or (
        header["version"]["major"] > 1
    ):
        data["t_amplifier"][
            indices["amplifier"] : (
                indices["amplifier"] + header["num_samples_per_data_block"]
            )
        ] = np.array(
            struct.unpack(
                "<" + "i" * header["num_samples_per_data_block"],
                fid.read(4 * header["num_samples_per_data_block"]),
            )
        )
    else:
        data["t_amplifier"][
            indices["amplifier"] : (
                indices["amplifier"] + header["num_samples_per_data_block"]
            )
        ] = np.array(
            struct.unpack(
                "<" + "I" * header["num_samples_per_data_block"],
                fid.read(4 * header["num_samples_per_data_block"]),
            )
        )

    if header["num_amplifier_channels"] > 0:
        tmp = np.fromfile(
            fid,
            dtype="uint16",
            count=header["num_samples_per_data_block"]
            * header["num_amplifier_channels"],
        )
        data["amplifier_data"][
            range(header["num_amplifier_channels"]),
            (indices["amplifier"]) : (
                indices["amplifier"] + header["num_samples_per_data_block"]
            ),
        ] = tmp.reshape(
            header["num_amplifier_channels"], header["num_samples_per_data_block"]
        )

    if header["num_aux_input_channels"] > 0:
        tmp = np.fromfile(
            fid,
            dtype="uint16",
            count=int(
                (header["num_samples_per_data_block"] / 4)
                * header["num_aux_input_channels"]
            ),
        )
        data["aux_input_data"][
            range(header["num_aux_input_channels"]),
            indices["aux_input"] : int(
                indices["aux_input"] + (header["num_samples_per_data_block"] / 4)
            ),
        ] = tmp.reshape(
            header["num_aux_input_channels"],
            int(header["num_samples_per_data_block"] / 4),
        )

    if header["num_supply_voltage_channels"] > 0:
        tmp = np.fromfile(
            fid, dtype="uint16", count=1 * header["num_supply_voltage_channels"]
        )
        data["supply_voltage_data"][
            range(header["num_supply_voltage_channels"]),
            indices["supply_voltage"] : (indices["supply_voltage"] + 1),
        ] = tmp.reshape(header["num_supply_voltage_channels"], 1)

    if header["num_temp_sensor_channels"] > 0:
        tmp = np.fromfile(
            fid, dtype="uint16", count=1 * header["num_temp_sensor_channels"]
        )
        data["temp_sensor_data"][
            range(header["num_temp_sensor_channels"]),
            indices["supply_voltage"] : (indices["supply_voltage"] + 1),
        ] = tmp.reshape(header["num_temp_sensor_channels"], 1)

    if header["num_board_adc_channels"] > 0:
        tmp = np.fromfile(
            fid,
            dtype="uint16",
            count=(header["num_samples_per_data_block"])
            * header["num_board_adc_channels"],
        )
        data["board_adc_data"][
            range(header["num_board_adc_channels"]),
            indices["board_adc"] : (
                indices["board_adc"] + header["num_samples_per_data_block"]
            ),
        ] = tmp.reshape(
            header["num_board_adc_channels"], header["num_samples_per_data_block"]
        )

    if header["num_board_dig_in_channels"] > 0:
        data["board_dig_in_raw"][
            indices["board_dig_in"] : (
                indices["board_dig_in"] + header["num_samples_per_data_block"]
            )
        ] = np.array(
            struct.unpack(
                "<" + "H" * header["num_samples_per_data_block"],
                fid.read(2 * header["num_samples_per_data_block"]),
            )
        )

    if header["num_board_dig_out_channels"] > 0:
        data["board_dig_out_raw"][
            indices["board_dig_out"] : (
                indices["board_dig_out"] + header["num_samples_per_data_block"]
            )
        ] = np.array(
            struct.unpack(
                "<" + "H" * header["num_samples_per_data_block"],
                fid.read(2 * header["num_samples_per_data_block"]),
            )
        )



#! /bin/env python
#
# Michael Gibson 27 April 2015

import math
import numpy as np


def notch_filter(input, fSample, fNotch, Bandwidth):
    """Implements a notch filter (e.g., for 50 or 60 Hz) on vector 'input'.

    fSample = sample rate of data (input Hz or Samples/sec)
    fNotch = filter notch frequency (input Hz)
    Bandwidth = notch 3-dB bandwidth (input Hz).  A bandwidth of 10 Hz is
    recommended for 50 or 60 Hz notch filters; narrower bandwidths lead to
    poor time-domain properties with an extended ringing response to
    transient disturbances.

    Example:  If neural data was sampled at 30 kSamples/sec
    and you wish to implement a 60 Hz notch filter:

    out = notch_filter(input, 30000, 60, 10);
    """

    tstep = 1.0 / fSample
    Fc = fNotch * tstep

    L = len(input)

    # Calculate IIR filter parameters
    d = math.exp(-2.0 * math.pi * (Bandwidth / 2.0) * tstep)
    b = (1.0 + d * d) * math.cos(2.0 * math.pi * Fc)
    a0 = 1.0
    a1 = -b
    a2 = d * d
    a = (1.0 + d * d) / 2.0
    b0 = 1.0
    b1 = -2.0 * math.cos(2.0 * math.pi * Fc)
    b2 = 1.0

    out = np.zeros(len(input))
    out[0] = input[0]
    out[1] = input[1]
    # (If filtering a continuous data stream, change out[0:1] to the
    #  previous final two values of out.)

    # Run filter
    for i in range(2, L):
        out[i] = (
            a * b2 * input[i - 2]
            + a * b1 * input[i - 1]
            + a * b0 * input[i]
            - a2 * out[i - 2]
            - a1 * out[i - 1]
        ) / a0

    return out



#! /bin/env python
#
# Michael Gibson 23 April 2015
# Modified Adrian Foy Sep 2018


def get_bytes_per_data_block(header):
    """Calculates the number of bytes in each 60 or 128 sample datablock."""

    # Each data block contains 60 or 128 amplifier samples.
    bytes_per_block = header["num_samples_per_data_block"] * 4  # timestamp data
    bytes_per_block = (
        bytes_per_block
        + header["num_samples_per_data_block"] * 2 * header["num_amplifier_channels"]
    )

    # Auxiliary inputs are sampled 4x slower than amplifiers
    bytes_per_block = (
        bytes_per_block
        + (header["num_samples_per_data_block"] / 4)
        * 2
        * header["num_aux_input_channels"]
    )

    # Supply voltage is sampled 60 or 128x slower than amplifiers
    bytes_per_block = bytes_per_block + 1 * 2 * header["num_supply_voltage_channels"]

    # Board analog inputs are sampled at same rate as amplifiers
    bytes_per_block = (
        bytes_per_block
        + header["num_samples_per_data_block"] * 2 * header["num_board_adc_channels"]
    )

    # Board digital inputs are sampled at same rate as amplifiers
    if header["num_board_dig_in_channels"] > 0:
        bytes_per_block = bytes_per_block + header["num_samples_per_data_block"] * 2

    # Board digital outputs are sampled at same rate as amplifiers
    if header["num_board_dig_out_channels"] > 0:
        bytes_per_block = bytes_per_block + header["num_samples_per_data_block"] * 2

    # Temp sensor is sampled 60 or 128x slower than amplifiers
    if header["num_temp_sensor_channels"] > 0:
        bytes_per_block = bytes_per_block + 1 * 2 * header["num_temp_sensor_channels"]

    return bytes_per_block



from spikeinterface.sorters import WaveClusSorter, IronClustSorter, Kilosort3Sorter
import pickle
import os
from pathlib import Path
import scipy
import spikeinterface

spikeinterface.__version__
import spikeinterface
import spikeinterface.extractors as se
# import spikeinterface.toolkit as st
import spikeinterface.sorters as ss
import spikeinterface.comparison as sc
import spikeinterface.widgets as sw
import numpy as np
import pandas as pd
from tqdm import tqdm
import numpy as np
from autosort_neuron.sorting import *


def detect_spike(
    trace0_car,
    thr_min=5,
    thr_max=30,
    distance=3,
    ch_max_simul_firing=3,
    wlen=5,
    prominence=10,
):
    noise_std_detect = np.median(abs(trace0_car) / 0.6745, axis=0)
    thr = thr_min * noise_std_detect
    thrmax = thr_max * noise_std_detect

    spikes = np.zeros(trace0_car.shape)
    if trace0_car.ndim > 1:
        for i in range(noise_std_detect.shape[0]):
            peaks, props = scipy.signal.find_peaks(
                -trace0_car[:, i],
                thr[i],
                distance=distance,
                wlen=wlen,
                prominence=prominence,
            )
            prominences = scipy.signal.peak_prominences(
                -trace0_car[:, i], peaks, wlen=7
            )[0]
            peaks = peaks[props["peak_heights"] > 10]
            prominences = prominences[props["peak_heights"] > 10]
            peaks = peaks[(prominences > 15)]

            spikes[peaks, i] = 1

        # larger value no more than thrmax
        points = trace0_car.shape[0]
        spike_coord = np.argwhere(spikes == 1)
        for i in range(spike_coord.shape[0]):
            near_start = spike_coord[i, 0] - 5
            near_end = spike_coord[i, 0] + 5
            if near_start < 0:
                near_start = 0
            if near_end >= points:
                near_end = points - 1
            if np.any(np.max(trace0_car[near_start:near_end, :], axis=0) >= thrmax):
                spikes[spike_coord[i, 0], spike_coord[i, 1]] = 0

        # no simultanous firing!!!!
        thres_cross = ch_max_simul_firing
        spikes[np.sum(spikes, axis=1) > thres_cross, :] = 0
    return spikes


def map_gt_annotation(detect_array, gt_array):
    gt_label_array1 = np.zeros((detect_array.shape[0],)) - 1

    for ind, i in enumerate(detect_array):
        f = 1
        indj = np.where(gt_array[:, 0] == i[0])[0]
        for j in indj:
            if gt_array[j, 1] == i[1]:
                f = 0
                break
        if f:
            indj = np.where(gt_array[:, 0] == i[0] - 1)[0]
            for j in indj:
                if gt_array[j, 1] == i[1]:
                    f = 0
                    break
        if f:
            indj = np.where(gt_array[:, 0] == i[0] + 1)[0]
            for j in indj:
                if gt_array[j, 1] == i[1]:
                    f = 0
                    break
        if f == 0:
            gt_label_array1[ind] = j

    return gt_label_array1


def save_obj(objt, name):
    with open(name + ".pkl", "wb") as f:
        pickle.dump(objt, f, pickle.HIGHEST_PROTOCOL)
        f.close()


def generate_autosort_input(
    date_id_all,
    raw_data_path,
    save_pth,
    day_pth,
    left_sample,
    right_sample,
    freq_min,
    freq_max,
    mesh_probe,
):
    all_save_folder_name = "_".join(date_id_all)

    for i in date_id_all:
        save_folder_name = i
        data_folder_all = f"./processed_data/Ephys_concat_{save_folder_name}/"
        if os.path.exists(data_folder_all) == False:
            _, _ = read_data_folder(
                data_folder_all,
                [i],
                raw_data_path,
                mesh_probe,
            )

        pth = f"./processed_data/Ephys_concat_{all_save_folder_name}/"
        extremum_channels_ids = pd.read_csv(
            pth + "mountainsort/extremum_channels_ids.csv", index_col=0
        )
        unit_list_all = {}
        for i, j in zip(
            extremum_channels_ids.index, extremum_channels_ids.values.flatten()
        ):
            unit_list_all[i] = j

        for set_time in np.arange(0, len(date_id_all)):

            date_id_all_i = date_id_all[set_time]
            if os.path.exists(save_pth + "/input/" + date_id_all_i + "/"):
                continue
            print("###", date_id_all_i)

            print("### 1. load raw data")
            recording_concat = spikeinterface.core.base.BaseExtractor.load_from_folder(
                day_pth + "Ephys_concat_" + date_id_all_i + "/"
            )
            recording_f = spikeinterface.preprocessing.bandpass_filter(
                recording_concat, freq_min=freq_min, freq_max=freq_max
            )
            recording_cmr = spikeinterface.preprocessing.common_reference(
                recording_f, reference="global", operator="average"
            )
            trace0_car = recording_cmr.get_traces(segment_index=0)

            print("### 2. detect spikes")
            spiketrain = {}
            all_spike_train = []
            spike_loc = []
            spikes = detect_spike(
                trace0_car,
                thr_min=3,
                thr_max=30,
                distance=3,
                ch_max_simul_firing=5,
                wlen=5,
                prominence=10,
            )
            for channel_num in range(trace0_car.shape[1]):
                spiketrain_loc = np.where(spikes[:, channel_num])[0]
                spiketrain[channel_num] = spiketrain_loc
                all_spike_train += list(spiketrain_loc)
                spike_loc += [channel_num] * len(spiketrain_loc)
            X_spiketrain_time = all_spike_train
            Y_spiketrain_id_final = spike_loc
            detect_array = np.array([X_spiketrain_time, Y_spiketrain_id_final]).T

            print("### 3. load ground truth")
            sorting = se.NpzSortingExtractor(
                pth + f"mountainsort/{date_id_all_i}/sorting/firings_merged.npz"
            )
            shank_id = sorting.unit_ids
            spike_train_all = []
            y_unit_id = []
            for i in tqdm(shank_id):
                add_unit = list(sorting.get_unit_spike_train(i))
                spike_train_all += add_unit
                y_unit_id += [i] * len(add_unit)
            gt_ch = [unit_list_all[i] for i in y_unit_id]
            gt_array = np.array([spike_train_all, gt_ch]).T

            print("### 4. map ground truth annotation")
            gt_label_array1 = map_gt_annotation(detect_array, gt_array)
            print(
                "---spike detection rate:",
                np.where(gt_label_array1 > -1)[0].shape[0] / gt_array.shape[0],
            )
            Y_spiketrain_id = np.zeros((detect_array.shape[0],)) - 1
            Y_spiketrain_id[np.where(gt_label_array1 > -1)[0]] = np.array(y_unit_id)[
                gt_label_array1[np.where(gt_label_array1 > -1)[0]].astype("int")
            ]

            print("### 4.5 add all gt")
            mapped_ind = gt_label_array1[np.where(gt_label_array1 > -1)[0]].astype(
                "int"
            )
            A = [i for i in np.arange(len(y_unit_id)) if i not in mapped_ind]

            X_spiketrain_time_train = list(X_spiketrain_time) + list(
                np.array(spike_train_all)[A]
            )
            Y_spiketrain_id_train = list(Y_spiketrain_id) + list(np.array(y_unit_id)[A])
            Y_spiketrain_id_final_train = list(Y_spiketrain_id_final) + list(
                np.array(gt_ch)[A]
            )

            print("### 5. find corresponding waveform")
            X_spiketrain_time = np.array(X_spiketrain_time)
            Y_spiketrain_id = np.array(Y_spiketrain_id)[
                X_spiketrain_time < trace0_car.shape[0] - (left_sample + right_sample)
            ]
            Y_spiketrain_id_final = np.array(Y_spiketrain_id_final)[
                X_spiketrain_time < trace0_car.shape[0] - (left_sample + right_sample)
            ]
            X_spiketrain_time = X_spiketrain_time[
                X_spiketrain_time < trace0_car.shape[0] - (left_sample + right_sample)
            ]
            for time_range in tqdm(np.arange(-left_sample, right_sample)):
                if time_range == -left_sample:
                    waveform = trace0_car[X_spiketrain_time + time_range, :]
                else:
                    waveform = np.dstack(
                        (waveform, trace0_car[X_spiketrain_time + time_range, :])
                    )

            print("### 6. save output")
            current_save_path = save_pth + "/input/" + date_id_all_i + "/test_data/"
            Path(current_save_path).mkdir(parents=True, exist_ok=True)
            save_obj(waveform, current_save_path + "/X_waveform")
            save_obj(Y_spiketrain_id, current_save_path + "/Y_spike_id")
            save_obj(Y_spiketrain_id_final, current_save_path + "/Y_spike_id_noise")
            save_obj(X_spiketrain_time, current_save_path + "/X_spiketrain_time")

            print("### 7. find corresponding waveform")
            X_spiketrain_time_train = np.array(X_spiketrain_time_train)
            Y_spiketrain_id_train = np.array(Y_spiketrain_id_train)[
                X_spiketrain_time_train
                < trace0_car.shape[0] - (left_sample + right_sample)
            ]
            Y_spiketrain_id_final_train = np.array(Y_spiketrain_id_final_train)[
                X_spiketrain_time_train
                < trace0_car.shape[0] - (left_sample + right_sample)
            ]
            X_spiketrain_time_train = X_spiketrain_time_train[
                X_spiketrain_time_train
                < trace0_car.shape[0] - (left_sample + right_sample)
            ]
            for time_range in tqdm(np.arange(-left_sample, right_sample)):
                if time_range == -left_sample:
                    waveform = trace0_car[X_spiketrain_time_train + time_range, :]
                else:
                    waveform = np.dstack(
                        (waveform, trace0_car[X_spiketrain_time_train + time_range, :])
                    )

            print("### 8. save output")
            current_save_path = save_pth + "/input/" + date_id_all_i + "/train_data/"
            Path(current_save_path).mkdir(parents=True, exist_ok=True)
            save_obj(waveform, current_save_path + "/X_waveform")
            save_obj(Y_spiketrain_id_train, current_save_path + "/Y_spike_id")
            save_obj(
                Y_spiketrain_id_final_train, current_save_path + "/Y_spike_id_noise"
            )
            save_obj(X_spiketrain_time_train, current_save_path + "/X_spiketrain_time")



#! /bin/env python
#
# Michael Gibson 23 April 2015
# Modified Adrian Foy Sep 2018

import sys, struct
from intanutil.qstring import read_qstring


def read_header(fid):
    """Reads the Intan File Format header from the given file."""

    # Check 'magic number' at beginning of file to make sure this is an Intan
    # Technologies RHD2000 data file.
    (magic_number,) = struct.unpack("<I", fid.read(4))
    if magic_number != int("c6912702", 16):
        raise Exception("Unrecognized file type.")

    header = {}
    # Read version number.
    version = {}
    (version["major"], version["minor"]) = struct.unpack("<hh", fid.read(4))
    header["version"] = version

    print("")
    print(
        "Reading Intan Technologies RHD2000 Data File, Version {}.{}".format(
            version["major"], version["minor"]
        )
    )
    print("")

    freq = {}

    # Read information of sampling rate and amplifier frequency settings.
    (header["sample_rate"],) = struct.unpack("<f", fid.read(4))
    (
        freq["dsp_enabled"],
        freq["actual_dsp_cutoff_frequency"],
        freq["actual_lower_bandwidth"],
        freq["actual_upper_bandwidth"],
        freq["desired_dsp_cutoff_frequency"],
        freq["desired_lower_bandwidth"],
        freq["desired_upper_bandwidth"],
    ) = struct.unpack("<hffffff", fid.read(26))

    # This tells us if a software 50/60 Hz notch filter was enabled during
    # the data acquisition.
    (notch_filter_mode,) = struct.unpack("<h", fid.read(2))
    header["notch_filter_frequency"] = 0
    if notch_filter_mode == 1:
        header["notch_filter_frequency"] = 50
    elif notch_filter_mode == 2:
        header["notch_filter_frequency"] = 60
    freq["notch_filter_frequency"] = header["notch_filter_frequency"]

    (
        freq["desired_impedance_test_frequency"],
        freq["actual_impedance_test_frequency"],
    ) = struct.unpack("<ff", fid.read(8))

    note1 = read_qstring(fid)
    note2 = read_qstring(fid)
    note3 = read_qstring(fid)
    header["notes"] = {"note1": note1, "note2": note2, "note3": note3}

    # If data file is from GUI v1.1 or later, see if temperature sensor data was saved.
    header["num_temp_sensor_channels"] = 0
    if (version["major"] == 1 and version["minor"] >= 1) or (version["major"] > 1):
        (header["num_temp_sensor_channels"],) = struct.unpack("<h", fid.read(2))

    # If data file is from GUI v1.3 or later, load eval board mode.
    header["eval_board_mode"] = 0
    if ((version["major"] == 1) and (version["minor"] >= 3)) or (version["major"] > 1):
        (header["eval_board_mode"],) = struct.unpack("<h", fid.read(2))

    header["num_samples_per_data_block"] = 60
    # If data file is from v2.0 or later (Intan Recording Controller), load name of digital reference channel
    if version["major"] > 1:
        header["reference_channel"] = read_qstring(fid)
        header["num_samples_per_data_block"] = 128

    # Place frequency-related information in data structure. (Note: much of this structure is set above)
    freq["amplifier_sample_rate"] = header["sample_rate"]
    freq["aux_input_sample_rate"] = header["sample_rate"] / 4
    freq["supply_voltage_sample_rate"] = (
        header["sample_rate"] / header["num_samples_per_data_block"]
    )
    freq["board_adc_sample_rate"] = header["sample_rate"]
    freq["board_dig_in_sample_rate"] = header["sample_rate"]

    header["frequency_parameters"] = freq

    # Create structure arrays for each type of data channel.
    header["spike_triggers"] = []
    header["amplifier_channels"] = []
    header["aux_input_channels"] = []
    header["supply_voltage_channels"] = []
    header["board_adc_channels"] = []
    header["board_dig_in_channels"] = []
    header["board_dig_out_channels"] = []

    # Read signal summary from data file header.

    (number_of_signal_groups,) = struct.unpack("<h", fid.read(2))
    print("n signal groups {}".format(number_of_signal_groups))

    for signal_group in range(1, number_of_signal_groups + 1):
        signal_group_name = read_qstring(fid)
        signal_group_prefix = read_qstring(fid)
        (
            signal_group_enabled,
            signal_group_num_channels,
            signal_group_num_amp_channels,
        ) = struct.unpack("<hhh", fid.read(6))

        if (signal_group_num_channels > 0) and (signal_group_enabled > 0):
            for signal_channel in range(0, signal_group_num_channels):
                new_channel = {
                    "port_name": signal_group_name,
                    "port_prefix": signal_group_prefix,
                    "port_number": signal_group,
                }
                new_channel["native_channel_name"] = read_qstring(fid)
                new_channel["custom_channel_name"] = read_qstring(fid)
                (
                    new_channel["native_order"],
                    new_channel["custom_order"],
                    signal_type,
                    channel_enabled,
                    new_channel["chip_channel"],
                    new_channel["board_stream"],
                ) = struct.unpack("<hhhhhh", fid.read(12))
                new_trigger_channel = {}
                (
                    new_trigger_channel["voltage_trigger_mode"],
                    new_trigger_channel["voltage_threshold"],
                    new_trigger_channel["digital_trigger_channel"],
                    new_trigger_channel["digital_edge_polarity"],
                ) = struct.unpack("<hhhh", fid.read(8))
                (
                    new_channel["electrode_impedance_magnitude"],
                    new_channel["electrode_impedance_phase"],
                ) = struct.unpack("<ff", fid.read(8))

                if channel_enabled:
                    if signal_type == 0:
                        header["amplifier_channels"].append(new_channel)
                        header["spike_triggers"].append(new_trigger_channel)
                    elif signal_type == 1:
                        header["aux_input_channels"].append(new_channel)
                    elif signal_type == 2:
                        header["supply_voltage_channels"].append(new_channel)
                    elif signal_type == 3:
                        header["board_adc_channels"].append(new_channel)
                    elif signal_type == 4:
                        header["board_dig_in_channels"].append(new_channel)
                    elif signal_type == 5:
                        header["board_dig_out_channels"].append(new_channel)
                    else:
                        raise Exception("Unknown channel type.")

    # Summarize contents of data file.
    header["num_amplifier_channels"] = len(header["amplifier_channels"])
    header["num_aux_input_channels"] = len(header["aux_input_channels"])
    header["num_supply_voltage_channels"] = len(header["supply_voltage_channels"])
    header["num_board_adc_channels"] = len(header["board_adc_channels"])
    header["num_board_dig_in_channels"] = len(header["board_dig_in_channels"])
    header["num_board_dig_out_channels"] = len(header["board_dig_out_channels"])

    return header


if __name__ == "__main__":
    h = read_header(open(sys.argv[1], "rb"))
    print(h)



import os
import pandas as pd
import torch
import torch.nn.functional as F
from torch.utils.data import random_split
import time
from waveform_loader import *
from tqdm import tqdm
from model import *
from sklearn.metrics import accuracy_score,f1_score
from pathlib import Path
from autosort_neuron.utils import seed_all
from sklearn.metrics import balanced_accuracy_score

#  use gpu if available
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
available_gpus = [torch.cuda.device(i) for i in range(torch.cuda.device_count())]
os.environ["CUDA_VISIBLE_DEVICES"] = "0" 
print(available_gpus)




def run(args):

    seed_all(args.seed_all)

    args.ch_num = args.group.shape[0]
    args.loc_dim = args.sensor_positions_all.shape[1]
    print(args)


    set_day_id_str = args.day_id_str[args.set_time]
    data_path = args.cluster_path+"input/"+set_day_id_str

    # if args.load_model == '0':
    save_dir = args.cluster_path+"model_save/"+"train_day"+set_day_id_str+'_'+str(args.seed_all)+'/train_weight/'
    Path(save_dir).mkdir(parents=True, exist_ok=True)


    set_channel_id = args.group
    sensor_positions = args.sensor_positions_all

    epochs = 20
    min_valid_loss = np.inf

    #### train AutoSort


    train_notpure_dataset = waveformLoader(data_path+'/train_data/',
                                            shank_channel=set_channel_id,
                                            sensor_positions=sensor_positions,
                                            )

        
    train, val = random_split(train_notpure_dataset, [int(len(train_notpure_dataset) * 0.8),
                                                        len(train_notpure_dataset) - int(
                                                            len(train_notpure_dataset) * 0.8)])
    train_loader = torch.utils.data.DataLoader(train, batch_size=512, shuffle=True, )
    val_loader = torch.utils.data.DataLoader(val, batch_size=512, shuffle=False, )


    autosort_model = AutoSort(
                                args.ch_num, 
                                args.samplepoints, 
                                args.loc_dim, 
                                device, 
                                train_notpure_dataset.keep_id,
                                save_dir,
                                pos_weight_noise=train_notpure_dataset.pos_weight_noise,
                                pos_weight_label=train_notpure_dataset.pos_weight_label
                                )
    if os.path.exists(autosort_model.save_model_path_2):
        autosort_model.load_model()
        autosort_model.to_device(device)

    else:
        training_log = {'epoch': [],
                        'training_loss1': [],
                        'training_loss2': [],
                        'training_loss3': [],
                        'validation_loss1': [],
                        'validation_loss2': [],
                        'validation_loss3': [],
                        'validation_acc_noise':[],
                        'validation_acc_label':[],
                        'validation_acc_noise_unbalanced': [],
                        'validation_acc_label_unbalanced': [],
                        'starttime_all':[],
                        'endtime_all':[]}

        for epoch in range(epochs):
            training_log['starttime_all'].append(time.perf_counter())
            training_log['epoch'].append(epoch + 1)
            print("epoch : {}/{}".format(epoch + 1, epochs))

            loss1 = 0
            loss2 = 0
            loss3 = 0
            autosort_model.train()
            autosort_model.bceloss.pos_weight = autosort_model.bceloss.pos_weight.to(device)
            autosort_model.bceloss_label.pos_weight = autosort_model.bceloss_label.pos_weight.to(device)
            for batch_features, classify_labels, labels, single_waveform,pred_loc in tqdm(train_loader):
                classify_labels = classify_labels.to(device)
                batch_features = batch_features.view(-1, args.samplepoints*args.ch_num).to(device)
                labels = labels.to(device)
                single_waveform  = single_waveform.to(device)
                pred_loc=torch.tensor(pred_loc).to(device)

                train_loss1, train_loss2, train_loss3, test = autosort_model.iter_model(batch_features, classify_labels, labels,single_waveform,pred_loc)

                loss1 += train_loss1
                loss2 += train_loss2
                if sum(test)>0:
                    loss3 += train_loss3

            loss1 = loss1 / len(train_loader)
            loss2 = loss2 / len(train_loader)
            loss3 = loss3 / len(train_loader)
            print("epoch : {}/{}, loss 1 = {:.6f}, loss 2 = {:.6f},, loss 3 = {:.6f}".format(epoch + 1, epochs, loss1,loss2,loss3))


            training_log['training_loss1'].append(loss1)
            training_log['training_loss2'].append(loss2)
            training_log['training_loss3'].append(loss3)


            valid_loss1 = 0.0
            valid_loss2 = 0.0
            valid_loss3 = 0.0

            gt_all = []
            pred_all = []
            gt_class_all = []
            pred_class_all = []
            autosort_model.eval()
            for data, classify_labels, labels, single_waveform,pred_loc in tqdm(val_loader):
                classify_labels = classify_labels.to(device)
                data = data.view(-1, args.samplepoints * args.ch_num).to(device)
                labels = labels.to(device)
                single_waveform  = single_waveform.to(device)
                pred_loc=torch.tensor(pred_loc).to(device)

                valid_loss1, valid_loss2, valid_loss3, gt, pred, gt_label_class, pred_class = autosort_model.iter_model_eval(data, classify_labels, labels,single_waveform,pred_loc)

                gt_all.append(gt.detach().cpu().numpy())
                pred_all.append(pred.detach().cpu().numpy())
                pred_class_all.append(pred_class.detach().cpu().numpy())
                gt_class_all.append(gt_label_class.detach().cpu().numpy())


            gt_all = np.concatenate(gt_all, axis=0)
            pred_all = np.concatenate(pred_all, axis=0)
            gt_class_all = np.concatenate(gt_class_all, axis=0)
            pred_class_all = np.concatenate(pred_class_all, axis=0)

            valid_loss1 = valid_loss1 / len(val_loader)
            valid_loss2 = valid_loss2 / len(val_loader)
            valid_loss3 = valid_loss3 / len(val_loader)
            valid_loss = valid_loss1 +valid_loss2 + valid_loss3
            print("epoch : {}/{}, val loss 1 = {:.6f}, loss 2 = {:.6f},loss 3 = {:.6f}".format(epoch + 1, epochs, valid_loss1, valid_loss2, valid_loss3))

            training_log['validation_loss1'].append(valid_loss1)
            training_log['validation_loss2'].append(valid_loss2)
            training_log['validation_loss3'].append(valid_loss3)
            training_log['validation_acc_noise'].append(accuracy_score(gt_all, pred_all))
            training_log['validation_acc_label'].append(f1_score(gt_class_all, pred_class_all,average='micro'))
            training_log['validation_acc_noise_unbalanced'].append(balanced_accuracy_score(gt_all, pred_all))
            training_log['validation_acc_label_unbalanced'].append(balanced_accuracy_score(gt_class_all, pred_class_all))
            training_log['endtime_all'].append(time.perf_counter())

            if min_valid_loss > valid_loss:
                print(f'Validation Loss Decreased({min_valid_loss:.6f}--->{valid_loss:.6f}) \t Saving The Model')
                min_valid_loss = valid_loss
                # Saving State Dict
                autosort_model.save_model()

        pd.DataFrame(training_log).to_csv(save_dir+'training_log.csv')


    #### test AutoSort
    test_log = {'train_time':[], 'timepoint':[],'noise_acc':[],'label_acc':[],'noise_acc_unbalanced':[],'label_acc_unbalanced':[],}

    for test_time in args.test_time:
        set_day_id_str = args.day_id_str[test_time]

        test_notpure_dataset = waveformLoader(data_path+'/test_data/',
                                        shank_channel = set_channel_id,
                                        sensor_positions=sensor_positions,
                                        Keep_id = train_notpure_dataset.keep_id,
                                        )


        test_notpure_loader = torch.utils.data.DataLoader(test_notpure_dataset,
        batch_size=512, shuffle=False,)

        gt_all = []
        pred_all = []
        gt_class_all = []
        pred_class_all = []
        autosort_model.eval()
        for data, classify_labels, labels, single_waveform,pred_loc in tqdm(test_notpure_loader):
            classify_labels = classify_labels.to(device)
            data = data.view(-1, args.samplepoints * args.ch_num).to(device)
            labels = labels.to(device)
            single_waveform  = single_waveform.to(device)
            pred_loc=torch.tensor(pred_loc).to(device)
            _,_,_, gt, pred, gt_label_class, pred_class = autosort_model.iter_model_eval(data, classify_labels, labels, single_waveform,pred_loc)
            gt_all.append(gt.detach().cpu().numpy())
            pred_all.append(pred.detach().cpu().numpy())
            pred_class_all.append(pred_class.detach().cpu().numpy())
            gt_class_all.append(gt_label_class.detach().cpu().numpy())


        gt_all = np.concatenate(gt_all, axis=0)
        pred_all = np.concatenate(pred_all, axis=0)
        gt_class_all = np.concatenate(gt_class_all, axis=0)
        pred_class_all = np.concatenate(pred_class_all, axis=0)

        test_log['train_time'].append(args.day_id_str[args.set_time])
        test_log['timepoint'].append(set_day_id_str)
        test_log['noise_acc'].append(accuracy_score(gt_all, pred_all))
        test_log['label_acc'].append(accuracy_score(gt_class_all, pred_class_all))
        test_log['noise_acc_unbalanced'].append(balanced_accuracy_score(gt_all, pred_all))
        test_log['label_acc_unbalanced'].append(balanced_accuracy_score(gt_class_all, pred_class_all))

    pd.DataFrame(test_log).to_csv(save_dir+'test_log.csv')




from setuptools import setup, find_packages

with open('README.md','r') as f:
    description = f.read()

setup(
    name='autosort_neuron',
    version='0.0.1.3',
    packages=find_packages(),
    install_requires=[
        'spikeinterface[full,widgets]',
        'torch',
        'pandas',
        'scikit-learn',
        'matplotlib',
        'scipy',
        'networkx',
    ],
    
    long_description=description,
    long_description_content_type='text/markdown',

)


import os
import torch
from torch import nn
import torch.nn.functional as F
import random
import numpy as np
import os
import tracemalloc


def display_top(snapshot, key_type="lineno", limit=3):
    snapshot = snapshot.filter_traces(
        (
            tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
            tracemalloc.Filter(False, "<unknown>"),
        )
    )
    top_stats = snapshot.statistics(key_type)

    print("Top %s lines" % limit)
    for index, stat in enumerate(top_stats[:limit], 1):
        frame = stat.traceback[0]
        # replace "/path/to/module/file.py" with "module/file.py"
        filename = os.sep.join(frame.filename.split(os.sep)[-2:])
        print(
            "#%s: %s:%s: %.1f KiB" % (index, filename, frame.lineno, stat.size / 1024)
        )
        line = linecache.getline(frame.filename, frame.lineno).strip()
        if line:
            print("    %s" % line)

    other = top_stats[limit:]
    if other:
        size = sum(stat.size for stat in other)
        print("%s other: %.1f KiB" % (len(other), size / 1024))
    total = sum(stat.size for stat in top_stats)
    print("Total allocated size: %.1f KiB" % (total / 1024))


def seed_all(seed_value, cuda_deterministic=False):
    print(
        "---------------------------------- SEED ALL ---------------------------------- "
    )
    print(
        f"                           Seed Num :   {seed_value}                                "
    )
    print(
        "---------------------------------- SEED ALL ---------------------------------- "
    )
    random.seed(seed_value)
    os.environ["PYTHONHASHSEED"] = str(seed_value)
    np.random.seed(seed_value)
    torch.manual_seed(seed_value)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed_value)
        torch.cuda.manual_seed_all(
            seed_value
        )  # Speed-reproducibility tradeoff https://pytorch.org/docs/stable/notes/randomness.html
        if cuda_deterministic:  # slower, more reproducible
            torch.backends.cudnn.deterministic = True
            torch.backends.cudnn.benchmark = False
        else:  # faster, less reproducible
            torch.backends.cudnn.deterministic = False
            torch.backends.cudnn.benchmark = True
